/*
 * Library interface to Linux x86 Performance-Monitoring Counters driver.
 *
 * Copyright (C) 1999  Mikael Pettersson
 */
#include "asm/perfctr.h"
#define CONFIG_PERFCTR
#include "linux/prctl.h"
#include <stdlib.h>
#include <errno.h>
#include "libperfctr.h"

static int init_done = 0;

#define CHECK_INIT() \
do { \
    if( !init_done && perfctr_init() < 0 ) \
	return -1; \
} while(0)

static int have_direct_access;
static struct perfctr_control current_control;
static struct perfctr_counters base; /* updated by kernel */

static __inline__ void rdtsc64(unsigned long long *u64)
{
    unsigned int *u32 = (unsigned int*)u64;
    __asm__ __volatile__("rdtsc " : "=a"(u32[0]), "=d"(u32[1]));
}

static __inline__ void rdpmc64(unsigned int ctr, unsigned long long *u64)
{
    unsigned int *u32 = (unsigned int*)u64;
    __asm__ __volatile__("rdpmc" : "=a"(u32[0]), "=d"(u32[1]) : "c"(ctr));
}

static int perfctr_init(void)
{
    struct perfctr_info info;

    if( prctl(PR_PERFCTR, PERFCTR_INFO, -1, &info) < 0 )
	return -1;
    /* This will be false for pre-MMX P5 (no rdpmc) and
     * WinChip processors (enabling perfctrs disables rdtsc).
     */
    have_direct_access = info.have_rdpmc && info.have_rdtsc;
    init_done = 1;
    return 0;
}

int perfctr_info(struct perfctr_info *info)
{
    CHECK_INIT();
    return prctl(PR_PERFCTR, PERFCTR_INFO, -1, info);
}

int perfctr_control_self(struct perfctr_control *control)
{
    CHECK_INIT();
    current_control = *control;
    current_control.user = &base;
    if( prctl(PR_PERFCTR, PERFCTR_CONTROL, PERFCTR_WHOM_SELF, &current_control) < 0 ) {
	current_control.status = 0;
	return -1;
    }
    return 0;
}

int perfctr_read_self(struct perfctr_state *state)
{
    CHECK_INIT();

    if( have_direct_access && current_control.status > 0 ) {
	unsigned long long tsc_start, tsc, pmc[2];
	do {
	    tsc_start = base.tsc_start;
	    rdtsc64(&tsc);
	    rdpmc64(0, &pmc[0]);
	    rdpmc64(1, &pmc[1]);
	    state->counters.tsc = base.tsc + (tsc - tsc_start);
	    state->counters.pmc[0] = base.pmc[0] + pmc[0];
	    state->counters.pmc[1] = base.pmc[1] + pmc[1];
	} while( tsc_start != base.tsc_start );
	state->counters.tsc_start = tsc_start;
	state->control = current_control;
	return 0;
    } else {
	if( prctl(PR_PERFCTR, PERFCTR_READ, PERFCTR_WHOM_SELF, state) < 0 ) {
	    if( errno == ENXIO ) {
		memset(state, 0, sizeof *state);
		return 0;
	    } else
		return -1;
	} else
	    return 0;
    }
}
